
/* Copyright 2012 Evan Miller */
/* All rights reserved. */

#pragma OPENCL EXTENSION cl_khr_fp64 : enable

#ifdef cl_khr_fp64
typedef double   fcl_fp;
typedef double4  fcl_fp4;
typedef double16 fcl_fp16;
typedef double4  fcl_double4;
typedef double16 fcl_double16;
#else
typedef float    fcl_fp;
typedef float4   fcl_fp4;
typedef float16  fcl_fp16;
typedef float4   fcl_double4;
typedef float16  fcl_double16;
#endif

#define FCL_1_OVER_SQRT_2_PI 0.3989422804014326f
#define FCL_1_OVER_SQRT_2_PI_D 0.3989422804014326

#define FCL_DIGAMMA_MIN_ARG_FOR_POLYNOMIAL_APPROX   8.5
#define FCL_DIGAMMA_MAX_ARG_FOR_LINEAR_APPROX       0.00001
#define FCL_DIGAMMA_D1                             -0.5772156649
#define FCL_DIGAMMA_S3                              0.08333333333
#define FCL_DIGAMMA_S4                              0.0083333333333
#define FCL_DIGAMMA_S5                              0.003968253968

#define FCL_TRIGAMMA_MIN_ARG_FOR_POLYNOMIAL_APPROX  5.
#define FCL_TRIGAMMA_MAX_ARG_FOR_LINEAR_APPROX      0.0001
#define FCL_TRIGAMMA_B2                             0.1666666667
#define FCL_TRIGAMMA_B4                            -0.03333333333
#define FCL_TRIGAMMA_B6                             0.02380952381
#define FCL_TRIGAMMA_B8                            -0.03333333333

#define fcl_slice_to_sliver(i) (4*i)

fcl_fp sum4(fcl_fp4 vec);
fcl_fp sum16(fcl_fp16 vec);
fcl_fp dot16(fcl_fp16 vec1, fcl_fp16 vec2);

fcl_fp4 select4(fcl_fp4 c, fcl_fp4 b, int4 a);
fcl_fp16 select16(fcl_fp16 c, fcl_fp16 b, int16 a);

fcl_fp fcl_logistic_cdf(fcl_fp alpha);
fcl_fp fcl_normal_cdf(fcl_fp alpha);

fcl_fp fcl_normal_pdf(fcl_fp alpha);
fcl_fp fcl_inverse_mills_ratio(fcl_fp alpha);

fcl_fp digamma(fcl_fp alpha);
fcl_fp16 digamma16(fcl_fp16 alpha);

fcl_fp trigamma(fcl_fp alpha);
fcl_fp16 trigamma16(fcl_fp16 alpha);

int4 convert_fp4_to_int4(fcl_fp4 res_fp);
int16 convert_fp16_to_int16(fcl_fp16 res_fp);

fcl_fp4 convert_int4_to_fp4(int4 res_i);
fcl_fp16 convert_int16_to_fp16(int16 res_i);

int4 int_array_elements4(__global int *array, int4 idx);
int16 int_array_elements16(__global int *array, int16 idx);
fcl_fp4 array_elements4(__global fcl_fp *array, int4 idx);
fcl_fp16 array_elements16(__global fcl_fp *array, int16 idx);

fcl_fp dot16(fcl_fp16 vec1, fcl_fp16 vec2) {
    fcl_fp result = 0.f;
    result += dot(vec1.lo.lo, vec2.lo.lo);
    result += dot(vec1.lo.hi, vec2.lo.hi);
    result += dot(vec1.hi.lo, vec2.hi.lo);
    result += dot(vec1.hi.hi, vec2.hi.hi);
    return result;
}

fcl_fp4 select4(fcl_fp4 c, fcl_fp4 b, int4 a) {
    fcl_fp4 res_fp;
    res_fp.s0 = a.s0 ? b.s0 : c.s0;
    res_fp.s1 = a.s1 ? b.s1 : c.s1;
    res_fp.s2 = a.s2 ? b.s2 : c.s2;
    res_fp.s3 = a.s3 ? b.s3 : c.s3;
    return res_fp;
}

fcl_fp16 select16(fcl_fp16 c, fcl_fp16 b, int16 a) {
    fcl_fp16 res_fp;
    res_fp.lo.lo = select4(c.lo.lo, b.lo.lo, a.lo.lo);
    res_fp.lo.hi = select4(c.lo.hi, b.lo.hi, a.lo.hi);
    res_fp.hi.lo = select4(c.hi.lo, b.hi.lo, a.hi.lo);
    res_fp.hi.hi = select4(c.hi.hi, b.hi.hi, a.hi.hi);
    return res_fp;
}

fcl_fp sum4(fcl_fp4 vec) {
    return vec.x + vec.y + vec.z + vec.w;
}

fcl_fp sum16(fcl_fp16 vec) {
    return sum4(vec.lo.lo) + sum4(vec.lo.hi) + sum4(vec.hi.lo) + sum4(vec.hi.hi);
}

fcl_fp fcl_logistic_cdf(fcl_fp alpha) {
    return 1.0 / (1.0 + exp(-alpha));
}

fcl_fp fcl_normal_cdf(fcl_fp alpha) {
    return 0.5 * erfc(-alpha * M_SQRT1_2);
}

fcl_fp fcl_normal_pdf(fcl_fp alpha) {
    return FCL_1_OVER_SQRT_2_PI_D * exp(-0.5 * alpha * alpha);
}

fcl_fp fcl_inverse_mills_ratio(fcl_fp alpha) {
    return fcl_normal_pdf(alpha) / fcl_normal_cdf(alpha);
}

fcl_fp digamma(fcl_fp alpha) {
    fcl_fp r;
    fcl_fp value = 0.0;
    fcl_fp y = alpha;

    if (alpha <= FCL_DIGAMMA_MAX_ARG_FOR_LINEAR_APPROX) {
        return FCL_DIGAMMA_D1 - 1.0 / y;
    }
    while (y < FCL_DIGAMMA_MIN_ARG_FOR_POLYNOMIAL_APPROX) {
        value -= 1.0 / y;
        y += 1.0;
    }
    value += log(y) - 0.5 / y;
    r = 1.0 / (y * y);

    return value - r*(FCL_DIGAMMA_S3 - r*(FCL_DIGAMMA_S4 - r*FCL_DIGAMMA_S5));
}

fcl_fp16 digamma16(fcl_fp16 alpha) {
    fcl_fp16 r;
    fcl_fp16 y = alpha;
    fcl_fp16 value = (fcl_fp16)(0.0);
    while (any(y < FCL_DIGAMMA_MIN_ARG_FOR_POLYNOMIAL_APPROX)) {
        value -= 1.0 / y;
        y += 1.0;
    }
    value += log(y) - 0.5 / y;
    r = 1.0 / (y * y);

    return select(value - r*(FCL_DIGAMMA_S3 - r*(FCL_DIGAMMA_S4 - r*FCL_DIGAMMA_S5)), 
            FCL_DIGAMMA_D1 - 1.0 / alpha, alpha <= FCL_DIGAMMA_MAX_ARG_FOR_LINEAR_APPROX);
}

fcl_fp trigamma(fcl_fp alpha) {
    fcl_fp value = 0.0;
    fcl_fp y;
    fcl_fp z = alpha;

    if (alpha <= FCL_TRIGAMMA_MAX_ARG_FOR_LINEAR_APPROX) {
        return 1.0 / (alpha * alpha);
    }

    while (z < FCL_TRIGAMMA_MIN_ARG_FOR_POLYNOMIAL_APPROX) {
        value += 1.0 / (z * z);
        z += 1.0;
    }

    y = 1.0 / (z * z);

    return value + 0.5 * y + ( 1.0
            + y * ( FCL_TRIGAMMA_B2
                + y * ( FCL_TRIGAMMA_B4
                    + y * ( FCL_TRIGAMMA_B6
                        + y * FCL_TRIGAMMA_B8 )))) / z;
}

fcl_fp16 trigamma16(fcl_fp16 alpha) {
    fcl_fp16 value = 0.0;
    fcl_fp16 y;
    fcl_fp16 z = alpha;

    while (any(z < FCL_TRIGAMMA_MIN_ARG_FOR_POLYNOMIAL_APPROX)) {
        value += 1.0 / (z * z);
        z += 1.0;
    }

    y = 1.0 / (z * z);

    return select(value + 0.5 * y + ( 1.0
            + y * ( FCL_TRIGAMMA_B2
                + y * ( FCL_TRIGAMMA_B4
                    + y * ( FCL_TRIGAMMA_B6
                        + y * FCL_TRIGAMMA_B8 )))) / z,
                        1. / (alpha * alpha), alpha <= FCL_TRIGAMMA_MAX_ARG_FOR_LINEAR_APPROX);
}

int4 convert_fp4_to_int4(fcl_fp4 res_fp) {
    int4 res_i;
    res_i.s0 = res_fp.s0;
    res_i.s1 = res_fp.s1;
    res_i.s2 = res_fp.s2;
    res_i.s3 = res_fp.s3;
    return res_i;
}

int16 convert_fp16_to_int16(fcl_fp16 res_fp) {
    int16 res_i;
    res_i.lo.lo = convert_fp4_to_int4(res_fp.lo.lo);
    res_i.lo.hi = convert_fp4_to_int4(res_fp.lo.hi);
    res_i.hi.lo = convert_fp4_to_int4(res_fp.hi.lo);
    res_i.hi.hi = convert_fp4_to_int4(res_fp.hi.hi);
    return res_i;
}

fcl_fp4 convert_int4_to_fp4(int4 res_i) {
    fcl_fp4 res_fp;
    res_fp.s0 = res_i.s0;
    res_fp.s1 = res_i.s1;
    res_fp.s2 = res_i.s2;
    res_fp.s3 = res_i.s3;
    return res_fp;
}

fcl_fp16 convert_int16_to_fp16(int16 res_i) {
    fcl_fp16 res_fp;
    res_fp.lo.lo = convert_int4_to_fp4(res_i.lo.lo);
    res_fp.lo.hi = convert_int4_to_fp4(res_i.lo.hi);
    res_fp.hi.lo = convert_int4_to_fp4(res_i.hi.lo);
    res_fp.hi.hi = convert_int4_to_fp4(res_i.hi.hi);
    return res_fp;
}

int4 int_array_elements4(__global int *array, int4 idx) {
    int4 res_i;
    res_i.s0 = array[idx.s0];
    res_i.s1 = array[idx.s1];
    res_i.s2 = array[idx.s2];
    res_i.s3 = array[idx.s3];
    return res_i;
}

int16 int_array_elements16(__global int *array, int16 idx) {
    int16 res_i;
    res_i.lo.lo = int_array_elements4(array, idx.lo.lo);
    res_i.lo.hi = int_array_elements4(array, idx.lo.hi);
    res_i.hi.lo = int_array_elements4(array, idx.hi.lo);
    res_i.hi.hi = int_array_elements4(array, idx.hi.hi);
    return res_i;
}

/* Copyright 2012 Evan Miller */
/* All rights reserved. */

__kernel void fcl_column_has_data(
    __global fcl_fp16 *data,
    __global fcl_fp16 *fweights,
    int stride,
    __global int  *independent_cols,
    __global int   *results_out
) {
    int k = get_global_id(0);
	__global fcl_fp16 *col = &data[independent_cols[k]*stride];
    int i;
    int has_data = 0;

    for (i=0; i<stride; i++) {
        if (any(col[i] * fweights[i] != 0.0)) {
            has_data = 1;
            break;
        }
    }

    results_out[k] = has_data;
}

__kernel void fcl_column_value_count(
    __global fcl_fp16 *data,
    __global fcl_fp16 *fweights,
    int index,
    int stride,
    __global fcl_fp *values_of_interest,
    __global fcl_fp *results_out
) {
    int j = get_global_id(0);
	__global fcl_fp16 *col = &data[index*stride];
    int i;
    fcl_fp count = 0.0;
    fcl_fp value_of_interest = values_of_interest[j];

    for (i=0; i<stride; i++) {
        count += sum16(select(0.0, fweights[i], col[i] == value_of_interest));
    }

    results_out[j] = count;
}

__kernel void fcl_compute_fweight_sum(
    __global fcl_fp16 *data,
    __global fcl_fp16 *fweights,
    int index,
    int stride,
    __global fcl_fp *sum_out) {
    int i;
    fcl_fp result = 0.f;

    for (i=0; i<stride; i++) {
        result += sum16(fweights[i]);
    }

    sum_out[0] = result;
}

__kernel void fcl_compute_integer_sum(
        __global int16 *data,
        __global fcl_fp16 *fweights,
        int index,
        int stride,
        __global fcl_fp *sum_out) {
    __global int16 *col = &data[index*stride];

    int i;
    fcl_fp result = 0.f;
    
    for (i=0; i<stride; i++) {
        result += dot16(fweights[i], convert_int16_to_fp16(col[i]));
    }
    
    sum_out[0] = result;
}

__kernel void fcl_compute_sum(
        __global fcl_fp16 *data,
        __global fcl_fp16 *fweights,
        int index,
        int stride,
        __global fcl_fp *sum_out) {
    __global fcl_fp16 *col = &data[index*stride];

    int i;
    fcl_fp result = 0.f;
    
    for (i=0; i<stride; i++) {
        result += dot16(fweights[i], col[i]);
    }
    
    sum_out[0] = result;
}

__kernel void fcl_compute_sum_of_squares(
        __global fcl_fp16 *data,
        __global fcl_fp16 *fweights,
        int index,
        int stride,
        __global fcl_fp *ss_out) {
    __global fcl_fp16 *col = &data[index*stride];

    int i;
    fcl_fp result = 0.f;

    for (i=0; i<stride; i++) {
        result += dot16(fweights[i], col[i] * col[i]);
    }

    ss_out[0] = result;
}

__kernel void fcl_compute_log_sum_of_squares(
        __global fcl_fp16 *data,
        __global fcl_fp16 *fweights,
        int index,
        int stride,
        __global fcl_fp *ss_out) {
    __global fcl_fp16 *col = &data[index*stride];

    int i;
    fcl_fp result = 0.f;

    for (i=0; i<stride; i++) {
        fcl_fp16 val = select(0.f, log(col[i]), col[i] > 0.f);
        result += dot16(fweights[i], val * val);
    }

    ss_out[0] = result;
}

__kernel void fcl_compute_centered_sum_of_squares(
        __global fcl_fp16 *data,
        __global fcl_fp16 *fweights,
        int index,
        int stride,
        __global fcl_fp *ss_out) {
    __global fcl_fp16 *col = &data[index*stride];

    int i;
    fcl_fp result = 0.f;
    fcl_fp mean = 0.f;
    fcl_fp total = 0.f;
    fcl_fp total_weight = 0.f;

    for (i=0; i<stride; i++) {
        total += dot16(fweights[i], col[i]);
        total_weight += sum16(fweights[i]);
    }

    mean = total / total_weight;

    for (i=0; i<stride; i++) {
        result += dot16(fweights[i], (col[i] - mean) * (col[i] - mean));
    }

    ss_out[0] = result;
}

__kernel void fcl_compute_log_centered_sum_of_squares(
        __global fcl_fp16 *data,
        __global fcl_fp16 *fweights,
        int index,
        int stride,
        __global fcl_fp *ss_out) {
    __global fcl_fp16 *col = &data[index*stride];

    int i;
    fcl_fp result = 0.f;
    fcl_fp mean = 0.f;
    fcl_fp total = 0.f;
    fcl_fp total_weight = 0.f;

    for (i=0; i<stride; i++) {
        total += dot16(fweights[i], select(0.f, log(col[i]), col[i] > 0.f));
        total_weight += sum16(fweights[i]);
    }

    mean = total / total_weight;

    for (i=0; i<stride; i++) {
        fcl_fp16 centered_val = select(0.f, log(col[i]) - mean, col[i] > 0.f);
        result += dot16(fweights[i], centered_val * centered_val);
    }

    ss_out[0] = result;
}

__kernel void fcl_compute_mean(
        __global fcl_fp16 *data,
        __global fcl_fp16 *fweights,
        int index,
        int stride,
        __global fcl_fp *mean_out) {
    __global fcl_fp16 *col = &data[index*stride];

    int i;
    fcl_fp result = 0.f;
    fcl_fp total_weight = 0.f;
    
    for (i=0; i<stride; i++) {
        result += dot16(fweights[i], col[i]);
        total_weight += sum16(fweights[i]);
    }
    
    mean_out[0] = result / total_weight;
}

__kernel void fcl_compute_log_mean(
        __global fcl_fp16 *data,
        __global fcl_fp16 *fweights,
        int index,
        int stride,
        __global fcl_fp *mean_out) {
    __global fcl_fp16 *col = &data[index*stride];

    int i;
    fcl_fp result = 0.f;
    fcl_fp total_weight = 0.f;
    
    for (i=0; i<stride; i++) {
        result += dot16(fweights[i], select(0.f, log(col[i]), col[i] > 0.f));
        total_weight += sum16(fweights[i]);
    }
    
    mean_out[0] = result / total_weight;
}

__kernel void fcl_compute_weighted_mean(
                                       __global fcl_fp16 *data,
                                       __global fcl_fp16 *fweights,
                                       int index,
                                       int stride,
                                       __global fcl_fp *mean_out,
                                       __global fcl_fp16 *aweights
                                       ) {
    __global fcl_fp16 *col = &data[index*stride];
    
    int i;
    fcl_fp result = 0.f;
    fcl_fp total_weight = 0.f;
    
    for (i=0; i<stride; i++) {
        fcl_fp16 aweight = aweights[i];
        fcl_fp16 fweight = fweights[i];
        fcl_fp16 val = col[i];
        result += sum16(fweight * aweight * val);
        total_weight += dot16(aweight, fweight);
    }
    
    mean_out[0] = result / total_weight;
}

__kernel void fcl_compute_weighted_log_mean(
                                       __global fcl_fp16 *data,
                                       __global fcl_fp16 *fweights,
                                       int index,
                                       int stride,
                                       __global fcl_fp *mean_out,
                                       __global fcl_fp16 *aweights
                                       ) {
    __global fcl_fp16 *col = &data[index*stride];
    
    int i;
    fcl_fp result = 0.f;
    fcl_fp total_weight = 0.f;
    
    for (i=0; i<stride; i++) {
        fcl_fp16 aweight = aweights[i];
        fcl_fp16 fweight = fweights[i];
        fcl_fp16 val = select(0.f, log(col[i]), col[i] > 0.f);
        result += sum16(fweight * aweight * val);
        total_weight += dot16(aweight, fweight);
    }
    
    mean_out[0] = result / total_weight;
}

__kernel void fcl_compute_panel_weighted_mean(
        __global fcl_fp4 *data,
        __global fcl_fp4 *fweights,
        int index,
        int slivers,
        __global fcl_fp *mean_out,
        __global fcl_fp *aweights,
        __global int *cluster_sizes,
        __global int *cluster_offsets,
        int panel_count,
        __global int *sliver2panel) 
{
    __global fcl_fp4 *col = &data[index*slivers];
    
    int i, t;
    fcl_fp result = 0.f;
    fcl_fp total_weight = 0.f;
    
    for (i=0; i<panel_count; i++) {
        int last = cluster_offsets[i] + cluster_sizes[i];
        fcl_fp weight = aweights[sliver2panel[i]];
        fcl_fp panel_weight = 0.f;
        for (t=cluster_offsets[i]; t<last; t++) {
            fcl_fp4 val = col[t];
            result += weight * dot(fweights[t], val);
            panel_weight += weight * sum4(fweights[t]);
        }
        total_weight += panel_weight;
    }
    
    mean_out[0] = result / total_weight;
}

__kernel void fcl_compute_panel_weighted_log_mean(
        __global fcl_fp4 *data,
        __global fcl_fp4 *fweights,
        int index,
        int slivers,
        __global fcl_fp *mean_out,
        __global fcl_fp *aweights,
        __global int *cluster_sizes,
        __global int *cluster_offsets,
        int panel_count,
        __global int *sliver2panel) 
{
    __global fcl_fp4 *col = &data[index*slivers];
    
    int i, t;
    fcl_fp result = 0.f;
    fcl_fp total_weight = 0.f;
    
    for (i=0; i<panel_count; i++) {
        int last = cluster_offsets[i] + cluster_sizes[i];
        fcl_fp weight = aweights[sliver2panel[i]];
        fcl_fp panel_weight = 0.f;
        for (t=cluster_offsets[i]; t<last; t++) {
            fcl_fp4 val = select(0.f, log(col[t]), col[t] > 0.f);
            result += weight * dot(fweights[t], val);
            panel_weight += weight * sum4(fweights[t]);
        }
        total_weight += panel_weight;
    }
    
    mean_out[0] = result / total_weight;
}

__kernel void fcl_compute_weighted_sum_of_squares(
                                       __global fcl_fp16 *data,
                                       __global fcl_fp16 *fweights,
                                       int index,
                                       int stride,
                                       __global fcl_fp *ss_out,
                                       __global fcl_fp16 *aweights
) {
    __global fcl_fp16 *col = &data[index*stride];
    fcl_fp result = 0.f;
    int i;

    for (i=0; i<stride; i++) {
        result += dot16(col[i] * col[i], select(0.f, 
                    fweights[i] / aweights[i], aweights[i] > 0.0));
    }

    ss_out[0] = result;
}

__kernel void fcl_compute_panel_weighted_sum_of_squares(
        __global fcl_fp4 *data,
        __global fcl_fp4 *fweights,
        int index,
        int slivers,
        __global fcl_fp *mean_out,
        __global fcl_fp *aweights,
        __global int *cluster_sizes,
        __global int *cluster_offsets,
        int panel_count,
        __global int *sliver2panel) 
{
    __global fcl_fp4 *col = &data[index*slivers];
    
    int i, t;
    fcl_fp result = 0.f;
    
    for (i=0; i<panel_count; i++) {
        int last = cluster_offsets[i] + cluster_sizes[i];
        fcl_fp weight = aweights[sliver2panel[i]];
        if (weight <= 0.0)
            continue;

        for (t=cluster_offsets[i]; t<last; t++) {
            fcl_fp4 val = col[t];
            result += dot(fweights[t] * val, val) / weight;
        }
    }
    
    mean_out[0] = result;
}
