
/* Copyright 2012 Evan Miller */
/* All rights reserved. */

#pragma OPENCL EXTENSION cl_khr_fp64 : enable

#ifdef cl_khr_fp64
typedef double   fcl_fp;
typedef double4  fcl_fp4;
typedef double16 fcl_fp16;
typedef double4  fcl_double4;
typedef double16 fcl_double16;
#else
typedef float    fcl_fp;
typedef float4   fcl_fp4;
typedef float16  fcl_fp16;
typedef float4   fcl_double4;
typedef float16  fcl_double16;
#endif

#define FCL_1_OVER_SQRT_2_PI 0.3989422804014326f
#define FCL_1_OVER_SQRT_2_PI_D 0.3989422804014326

#define FCL_DIGAMMA_MIN_ARG_FOR_POLYNOMIAL_APPROX   8.5
#define FCL_DIGAMMA_MAX_ARG_FOR_LINEAR_APPROX       0.00001
#define FCL_DIGAMMA_D1                             -0.5772156649
#define FCL_DIGAMMA_S3                              0.08333333333
#define FCL_DIGAMMA_S4                              0.0083333333333
#define FCL_DIGAMMA_S5                              0.003968253968

#define FCL_TRIGAMMA_MIN_ARG_FOR_POLYNOMIAL_APPROX  5.
#define FCL_TRIGAMMA_MAX_ARG_FOR_LINEAR_APPROX      0.0001
#define FCL_TRIGAMMA_B2                             0.1666666667
#define FCL_TRIGAMMA_B4                            -0.03333333333
#define FCL_TRIGAMMA_B6                             0.02380952381
#define FCL_TRIGAMMA_B8                            -0.03333333333

#define fcl_slice_to_sliver(i) (4*i)

fcl_fp sum4(fcl_fp4 vec);
fcl_fp sum16(fcl_fp16 vec);
fcl_fp dot16(fcl_fp16 vec1, fcl_fp16 vec2);

fcl_fp4 select4(fcl_fp4 c, fcl_fp4 b, int4 a);
fcl_fp16 select16(fcl_fp16 c, fcl_fp16 b, int16 a);

fcl_fp fcl_logistic_cdf(fcl_fp alpha);
fcl_fp fcl_normal_cdf(fcl_fp alpha);

fcl_fp fcl_normal_pdf(fcl_fp alpha);
fcl_fp fcl_inverse_mills_ratio(fcl_fp alpha);

fcl_fp digamma(fcl_fp alpha);
fcl_fp16 digamma16(fcl_fp16 alpha);

fcl_fp trigamma(fcl_fp alpha);
fcl_fp16 trigamma16(fcl_fp16 alpha);

int4 convert_fp4_to_int4(fcl_fp4 res_fp);
int16 convert_fp16_to_int16(fcl_fp16 res_fp);

fcl_fp4 convert_int4_to_fp4(int4 res_i);
fcl_fp16 convert_int16_to_fp16(int16 res_i);

int4 int_array_elements4(__global int *array, int4 idx);
int16 int_array_elements16(__global int *array, int16 idx);
fcl_fp4 array_elements4(__global fcl_fp *array, int4 idx);
fcl_fp16 array_elements16(__global fcl_fp *array, int16 idx);

fcl_fp dot16(fcl_fp16 vec1, fcl_fp16 vec2) {
    fcl_fp result = 0.f;
    result += dot(vec1.lo.lo, vec2.lo.lo);
    result += dot(vec1.lo.hi, vec2.lo.hi);
    result += dot(vec1.hi.lo, vec2.hi.lo);
    result += dot(vec1.hi.hi, vec2.hi.hi);
    return result;
}

fcl_fp4 select4(fcl_fp4 c, fcl_fp4 b, int4 a) {
    fcl_fp4 res_fp;
    res_fp.s0 = a.s0 ? b.s0 : c.s0;
    res_fp.s1 = a.s1 ? b.s1 : c.s1;
    res_fp.s2 = a.s2 ? b.s2 : c.s2;
    res_fp.s3 = a.s3 ? b.s3 : c.s3;
    return res_fp;
}

fcl_fp16 select16(fcl_fp16 c, fcl_fp16 b, int16 a) {
    fcl_fp16 res_fp;
    res_fp.lo.lo = select4(c.lo.lo, b.lo.lo, a.lo.lo);
    res_fp.lo.hi = select4(c.lo.hi, b.lo.hi, a.lo.hi);
    res_fp.hi.lo = select4(c.hi.lo, b.hi.lo, a.hi.lo);
    res_fp.hi.hi = select4(c.hi.hi, b.hi.hi, a.hi.hi);
    return res_fp;
}

fcl_fp sum4(fcl_fp4 vec) {
    return vec.x + vec.y + vec.z + vec.w;
}

fcl_fp sum16(fcl_fp16 vec) {
    return sum4(vec.lo.lo) + sum4(vec.lo.hi) + sum4(vec.hi.lo) + sum4(vec.hi.hi);
}

fcl_fp fcl_logistic_cdf(fcl_fp alpha) {
    return 1.0 / (1.0 + exp(-alpha));
}

fcl_fp fcl_normal_cdf(fcl_fp alpha) {
    return 0.5 * erfc(-alpha * M_SQRT1_2);
}

fcl_fp fcl_normal_pdf(fcl_fp alpha) {
    return FCL_1_OVER_SQRT_2_PI_D * exp(-0.5 * alpha * alpha);
}

fcl_fp fcl_inverse_mills_ratio(fcl_fp alpha) {
    return fcl_normal_pdf(alpha) / fcl_normal_cdf(alpha);
}

fcl_fp digamma(fcl_fp alpha) {
    fcl_fp r;
    fcl_fp value = 0.0;
    fcl_fp y = alpha;

    if (alpha <= FCL_DIGAMMA_MAX_ARG_FOR_LINEAR_APPROX) {
        return FCL_DIGAMMA_D1 - 1.0 / y;
    }
    while (y < FCL_DIGAMMA_MIN_ARG_FOR_POLYNOMIAL_APPROX) {
        value -= 1.0 / y;
        y += 1.0;
    }
    value += log(y) - 0.5 / y;
    r = 1.0 / (y * y);

    return value - r*(FCL_DIGAMMA_S3 - r*(FCL_DIGAMMA_S4 - r*FCL_DIGAMMA_S5));
}

fcl_fp16 digamma16(fcl_fp16 alpha) {
    fcl_fp16 r;
    fcl_fp16 y = alpha;
    fcl_fp16 value = (fcl_fp16)(0.0);
    while (any(y < FCL_DIGAMMA_MIN_ARG_FOR_POLYNOMIAL_APPROX)) {
        value -= 1.0 / y;
        y += 1.0;
    }
    value += log(y) - 0.5 / y;
    r = 1.0 / (y * y);

    return select(value - r*(FCL_DIGAMMA_S3 - r*(FCL_DIGAMMA_S4 - r*FCL_DIGAMMA_S5)), 
            FCL_DIGAMMA_D1 - 1.0 / alpha, alpha <= FCL_DIGAMMA_MAX_ARG_FOR_LINEAR_APPROX);
}

fcl_fp trigamma(fcl_fp alpha) {
    fcl_fp value = 0.0;
    fcl_fp y;
    fcl_fp z = alpha;

    if (alpha <= FCL_TRIGAMMA_MAX_ARG_FOR_LINEAR_APPROX) {
        return 1.0 / (alpha * alpha);
    }

    while (z < FCL_TRIGAMMA_MIN_ARG_FOR_POLYNOMIAL_APPROX) {
        value += 1.0 / (z * z);
        z += 1.0;
    }

    y = 1.0 / (z * z);

    return value + 0.5 * y + ( 1.0
            + y * ( FCL_TRIGAMMA_B2
                + y * ( FCL_TRIGAMMA_B4
                    + y * ( FCL_TRIGAMMA_B6
                        + y * FCL_TRIGAMMA_B8 )))) / z;
}

fcl_fp16 trigamma16(fcl_fp16 alpha) {
    fcl_fp16 value = 0.0;
    fcl_fp16 y;
    fcl_fp16 z = alpha;

    while (any(z < FCL_TRIGAMMA_MIN_ARG_FOR_POLYNOMIAL_APPROX)) {
        value += 1.0 / (z * z);
        z += 1.0;
    }

    y = 1.0 / (z * z);

    return select(value + 0.5 * y + ( 1.0
            + y * ( FCL_TRIGAMMA_B2
                + y * ( FCL_TRIGAMMA_B4
                    + y * ( FCL_TRIGAMMA_B6
                        + y * FCL_TRIGAMMA_B8 )))) / z,
                        1. / (alpha * alpha), alpha <= FCL_TRIGAMMA_MAX_ARG_FOR_LINEAR_APPROX);
}

int4 convert_fp4_to_int4(fcl_fp4 res_fp) {
    int4 res_i;
    res_i.s0 = res_fp.s0;
    res_i.s1 = res_fp.s1;
    res_i.s2 = res_fp.s2;
    res_i.s3 = res_fp.s3;
    return res_i;
}

int16 convert_fp16_to_int16(fcl_fp16 res_fp) {
    int16 res_i;
    res_i.lo.lo = convert_fp4_to_int4(res_fp.lo.lo);
    res_i.lo.hi = convert_fp4_to_int4(res_fp.lo.hi);
    res_i.hi.lo = convert_fp4_to_int4(res_fp.hi.lo);
    res_i.hi.hi = convert_fp4_to_int4(res_fp.hi.hi);
    return res_i;
}

fcl_fp4 convert_int4_to_fp4(int4 res_i) {
    fcl_fp4 res_fp;
    res_fp.s0 = res_i.s0;
    res_fp.s1 = res_i.s1;
    res_fp.s2 = res_i.s2;
    res_fp.s3 = res_i.s3;
    return res_fp;
}

fcl_fp16 convert_int16_to_fp16(int16 res_i) {
    fcl_fp16 res_fp;
    res_fp.lo.lo = convert_int4_to_fp4(res_i.lo.lo);
    res_fp.lo.hi = convert_int4_to_fp4(res_i.lo.hi);
    res_fp.hi.lo = convert_int4_to_fp4(res_i.hi.lo);
    res_fp.hi.hi = convert_int4_to_fp4(res_i.hi.hi);
    return res_fp;
}

int4 int_array_elements4(__global int *array, int4 idx) {
    int4 res_i;
    res_i.s0 = array[idx.s0];
    res_i.s1 = array[idx.s1];
    res_i.s2 = array[idx.s2];
    res_i.s3 = array[idx.s3];
    return res_i;
}

int16 int_array_elements16(__global int *array, int16 idx) {
    int16 res_i;
    res_i.lo.lo = int_array_elements4(array, idx.lo.lo);
    res_i.lo.hi = int_array_elements4(array, idx.lo.hi);
    res_i.hi.lo = int_array_elements4(array, idx.hi.lo);
    res_i.hi.hi = int_array_elements4(array, idx.hi.hi);
    return res_i;
}

/* Copyright 2012 Evan Miller */
/* All rights reserved. */

/* See Greene and Hensher, _Modeling Ordered Choices_, pp. 173-175 */

__kernel void fcl_prefix_ordered_probit(
        __global fcl_fp *data,
        int dependent_col,
        __global fcl_fp *fitted,
        __global fcl_fp *core_jacobian_prefix_out,
        __global fcl_fp *core_hessian_prefix_out,

        __global fcl_fp *estimates,
        int independent_cols_count,

        __global int *dep_vals_reverse_lookup,
        int dep_vals_count,

        __global fcl_fp *nuisance_jacobian_prefix_out,
        __global fcl_fp *nuisance_jacobian1_prefix_out,

        __global fcl_fp *cross_hessian_prefix_out,
        __global fcl_fp *cross_hessian1_prefix_out,

        __global fcl_fp *nuisance_hessian_prefix_out,
        __global fcl_fp *nuisance_hessian1_prefix_out,
        __global fcl_fp *nuisance_hessian_cross_prefix_out,

        __global fcl_fp *log_likelihood_out) 
{
    int i = get_global_id(0);
    int stride = get_global_size(0);

    int j;
    int actual = dep_vals_reverse_lookup[(int)data[dependent_col*stride+i]];
    __global fcl_fp *nuisance = estimates + independent_cols_count;

    fcl_fp xb = fitted[i];
    fcl_fp mu_j = nuisance[actual + 1];
    fcl_fp mu_j1 = nuisance[actual];

    fcl_fp d0 = mu_j - xb;
    fcl_fp d1 = mu_j1 - xb;

    fcl_fp Phi =  (actual == dep_vals_count - 1) ? 1.f : fcl_normal_cdf(d0);
    fcl_fp Phi1 = (actual == 0)                  ? 0.f : fcl_normal_cdf(d1);
    fcl_fp P = Phi - Phi1;
    
    fcl_fp f =       (actual == dep_vals_count - 1) ? 0.f : fcl_normal_pdf(d0);
    fcl_fp f1 =      (actual == 0)                  ? 0.f : fcl_normal_pdf(d1);
    fcl_fp f_prime = (actual == dep_vals_count - 1) ? 0.f : - d0 * f;
    fcl_fp f_prime1 = (actual == 0)                 ? 0.f : - d1 * f1;

    core_jacobian_prefix_out[i] = -(f - f1) / P;
    nuisance_jacobian_prefix_out[i] = f / P;
    nuisance_jacobian1_prefix_out[i] = - f1 / P;
    
    core_hessian_prefix_out[i] = (f_prime - f_prime1) / P - (f - f1) / P * (f - f1) / P;
    cross_hessian_prefix_out[i] = -(f_prime / P - (f - f1) / P * f / P);
    cross_hessian1_prefix_out[i] = f_prime1 / P + (f - f1) / P * (-f1) / P;
    nuisance_hessian_prefix_out[i] = f_prime / P - f / P * f / P;
    nuisance_hessian1_prefix_out[i] = -f_prime1 / P - f1 / P * f1 / P;
    nuisance_hessian_cross_prefix_out[i] = f / P * f1 / P;
    log_likelihood_out[i] = log(P);
}

__kernel void fcl_predict_ordered_probit(
        int independent_cols_count,
        __global fcl_fp *estimates,
        __global fcl_fp *fitted,
        __global fcl_fp *predicted_out
        ) {
    int j = get_global_id(0);
    int dep_vals_count = get_global_size(0);
    int i = get_global_id(1);
    int slices = get_global_size(1);

    __global fcl_fp *nuisance = estimates + independent_cols_count;
    
    fcl_fp Phi  = (j == dep_vals_count-1) ? 1.f : fcl_normal_cdf(nuisance[j+1]-fitted[i]);
    fcl_fp Phi1 = (j == 0)                ? 0.f : fcl_normal_cdf(nuisance[j]  -fitted[i]);

    predicted_out[j*slices+i] = Phi - Phi1;
}

__kernel void fcl_prefix_ordered_logit(
        __global fcl_fp *data,
        int dependent_col,
        __global fcl_fp *fitted,
        __global fcl_fp *core_jacobian_prefix_out,
        __global fcl_fp *core_hessian_prefix_out,

        __global fcl_fp *estimates,
        int independent_cols_count,

        __global int *dep_vals_reverse_lookup,
        int dep_vals_count,

        __global fcl_fp *nuisance_jacobian_prefix_out,
        __global fcl_fp *nuisance_jacobian1_prefix_out,

        __global fcl_fp *cross_hessian_prefix_out,
        __global fcl_fp *cross_hessian1_prefix_out,

        __global fcl_fp *nuisance_hessian_prefix_out,
        __global fcl_fp *nuisance_hessian1_prefix_out,
        __global fcl_fp *nuisance_hessian_cross_prefix_out,

        __global fcl_fp *log_likelihood_out) 
{
    int i = get_global_id(0);
    int stride = get_global_size(0);
    
    int j;
    int actual = dep_vals_reverse_lookup[(int)data[dependent_col*stride+i]];
    __global fcl_fp *nuisance = estimates + independent_cols_count;

    fcl_fp xb = fitted[i];
    
    fcl_fp mu_j = nuisance[actual + 1];
    fcl_fp mu_j1 = nuisance[actual];

    fcl_fp d0 = mu_j - xb;
    fcl_fp d1 = mu_j1 - xb;

    fcl_fp lambda = fcl_logistic_cdf(d0);
    fcl_fp lambda1 = fcl_logistic_cdf(d1);
    fcl_fp Phi =  (actual == dep_vals_count - 1) ? 1.f : lambda;
    fcl_fp Phi1 = (actual == 0)                  ? 0.f : lambda1;
    fcl_fp P = Phi - Phi1;

    fcl_fp f =        (actual == dep_vals_count - 1) ? 0.f : lambda  * (1.f - lambda);
    fcl_fp f1 =       (actual == 0)                  ? 0.f : lambda1 * (1.f - lambda1);
    fcl_fp f_prime =  (actual == dep_vals_count - 1) ? 0.f : lambda *  (1.f - lambda)  * (1.f - 2.f * lambda);
    fcl_fp f_prime1 = (actual == 0)                  ? 0.f : lambda1 * (1.f - lambda1) * (1.f - 2.f * lambda1);
    
    core_jacobian_prefix_out[i] = -(f - f1) / P;
    nuisance_jacobian_prefix_out[i] = f / P;
    nuisance_jacobian1_prefix_out[i] = - f1 / P;
    
    core_hessian_prefix_out[i] = (f_prime - f_prime1) / P - (f - f1) / P * (f - f1) / P;
    cross_hessian_prefix_out[i] = -(f_prime / P - (f - f1) / P * f / P);
    cross_hessian1_prefix_out[i] = f_prime1 / P + (f - f1) / P * (-f1) / P;
    nuisance_hessian_prefix_out[i] = f_prime / P - f / P * f / P;
    nuisance_hessian1_prefix_out[i] = -f_prime1 / P - f1 / P * f1 / P;
    nuisance_hessian_cross_prefix_out[i] = f / P * f1 / P;
    log_likelihood_out[i] = log(P);
}

__kernel void fcl_predict_ordered_logit(
        int independent_cols_count,
        __global fcl_fp *estimates,
        __global fcl_fp *fitted,
        __global fcl_fp *predicted_out
        ) {
    int j = get_global_id(0);
    int dep_vals_count = get_global_size(0);
    int i = get_global_id(1);
    int slices = get_global_size(1);

    __global fcl_fp *nuisance = estimates + independent_cols_count;
    
    fcl_fp Phi  = (j == dep_vals_count-1) ? 1.f : fcl_logistic_cdf(nuisance[j+1]-fitted[i]);
    fcl_fp Phi1 = (j == 0)                ? 0.f : fcl_logistic_cdf(nuisance[j]  -fitted[i]);

    predicted_out[j*slices+i] = Phi - Phi1;
}
